/*******************************************************************************
 * Copyright 2013 Francesco Cina'
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package com.jporm.query.crud.executor;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;

import com.jporm.exception.OrmOptimisticLockException;
import com.jporm.mapper.OrmClassTool;
import com.jporm.mapper.ServiceCatalog;
import com.jporm.mapper.clazz.ClassField;
import com.jporm.mapper.clazz.ClassMap;
import com.jporm.mapper.relation.RelationInnerFK;
import com.jporm.mapper.relation.RelationOuterFK;
import com.jporm.persistor.OrmPersistor;
import com.jporm.persistor.reflection.SetManipulator;
import com.jporm.query.find.FindWhere;
import com.jporm.query.save.AColumnValueGenerator;
import com.jporm.query.save.ColumnValueGeneratorFactory;
import com.jporm.query.update.CustomUpdateQuery;
import com.jporm.query.update.CustomUpdateSet;
import com.jporm.query.update.CustomUpdateWhere;
import com.jporm.session.GeneratedKeyReader;
import com.jporm.session.SqlExecutor;

/**
 * <class_description>
 * <p><b>notes</b>:
 * <p>ON : Mar 13, 2013
 *
 * @author  - Francesco Cina
 * @version $Revision
 */
public class OrmCRUDQueryExecutorSaveOrUpdateImpl implements OrmCRUDQueryExecutorSaveOrUpdate {

    private final ServiceCatalog serviceCatalog;

    public OrmCRUDQueryExecutorSaveOrUpdateImpl(final ServiceCatalog serviceCatalog) {
        this.serviceCatalog = serviceCatalog;
    }

    @Override
    public <BEAN> BEAN update(final BEAN bean, final Class<BEAN> clazz, final boolean cascade, final SaveOrUpdateType saveOrUpdateType, final int queryTimeout) {
        OrmClassTool<BEAN> ormClassTool = serviceCatalog.getOrmClassTool(clazz);
        saveOrUpdateInnerRelations(cascade, bean, ormClassTool, saveOrUpdateType);

        final OrmPersistor<BEAN> persistor = ormClassTool.getOrmPersistor();

        String[] pkAndVersionFieldNames = ormClassTool.getClassMap().getPrimaryKeyAndVersionColumnJavaNames();

        // CHECK IF OBJECT HAS A 'VERSION' FIELD AND THE DATA MUST BE LOCKED BEFORE UPDATE
        if (persistor.isVersionableWithLock()) {
            FindWhere<BEAN> query = (FindWhere<BEAN>) serviceCatalog.getSession().findQuery(bean.getClass())
                    .lockMode(persistor.getVersionableLockMode()).where();
            Object[] values = persistor.getPropertyValues(pkAndVersionFieldNames, bean);
            for (int i = 0; i < pkAndVersionFieldNames.length; i++) {
                query.eq(pkAndVersionFieldNames[i], values[i]);
            }
            if (query.getRowCount() == 0) {
                throw new OrmOptimisticLockException(
                        "The bean of class [" + bean.getClass() + "] cannot be updated. Version in the DB is not the expected one."); //$NON-NLS-1$ //$NON-NLS-2$
            }
        }

        CustomUpdateQuery updateQuery = serviceCatalog.getSession().updateQuery(bean.getClass()).queryTimeout(queryTimeout);

        CustomUpdateWhere updateQueryWhere = updateQuery.where();
        Object[] pkAndVersionValues = persistor.getPropertyValues(pkAndVersionFieldNames, bean);
        for (int i = 0; i < pkAndVersionFieldNames.length; i++) {
            updateQueryWhere.eq(pkAndVersionFieldNames[i], pkAndVersionValues[i]);
        }

        persistor.increaseVersion(bean, false);

        CustomUpdateSet updateQuerySet = updateQuery.set();
        String[] notPks = ormClassTool.getClassMap().getNotPrimaryKeyColumnJavaNames();
        Object[] notPkValues = persistor.getPropertyValues(notPks, bean);
        for (int i = 0; i < notPks.length; i++) {
            updateQuerySet.eq(notPks[i], notPkValues[i]);
        }

        if (updateQuery.now() == 0) {
            throw new OrmOptimisticLockException(
                    "The bean of class [" + bean.getClass() + "] cannot be updated. Version in the DB is not the expected one or the ID of the bean is associated with and existing bean."); //$NON-NLS-1$ //$NON-NLS-2$
        }

        saveOrUpdateOuterRelations(cascade, bean, ormClassTool, saveOrUpdateType);
        return bean;
    }

    @Override
    public <BEAN> BEAN save(final BEAN bean, final Class<BEAN> clazz, final boolean cascade, final SaveOrUpdateType saveOrUpdateType, final int queryTimeout) {
        final OrmClassTool<BEAN> ormClassTool = serviceCatalog.getOrmClassTool(clazz);
        saveOrUpdateInnerRelations(cascade, bean, ormClassTool, saveOrUpdateType);

        final OrmPersistor<BEAN> persistor = ormClassTool.getOrmPersistor();
        final SqlExecutor sqlExec = serviceCatalog.getSession().sqlExecutor();
        sqlExec.setQueryTimeout(queryTimeout);

        //CHECK IF OBJECT HAS A 'VERSION' FIELD and increase it
        persistor.increaseVersion(bean, true);
        boolean useGenerator = ormClassTool.getOrmPersistor().useGenerators(bean);
        String sql = generateSaveQuery(useGenerator, ormClassTool.getClassMap());
        if (!useGenerator) {
            String[] keys = ormClassTool.getClassMap().getAllColumnJavaNames();
            Object[] values = persistor.getPropertyValues(keys, bean);
            sqlExec.update(sql, values);
        } else {
            final GeneratedKeyReader generatedKeyExtractor = new GeneratedKeyReader() {

                @Override
                public void read(final ResultSet generatedKeyResultSet) throws SQLException {
                    if (generatedKeyResultSet.next()) {
                        persistor.updateGeneratedValues(generatedKeyResultSet, bean);
                    }
                }

                @Override
                public String[] generatedColumnNames() {
                    return ormClassTool.getClassMap().getAllGeneratedColumnDBNames();
                }
            };
            String[] keys = ormClassTool.getClassMap().getAllNotGeneratedColumnJavaNames();
            Object[] values = persistor.getPropertyValues(keys, bean);
            sqlExec.update(sql, generatedKeyExtractor, values);
        }
        saveOrUpdateOuterRelations(cascade, bean, ormClassTool, saveOrUpdateType);
        return bean;

    }

    @SuppressWarnings("unchecked")
    protected final <RELATION, BEAN> void saveOrUpdateInnerRelations(final boolean cascade, final BEAN bean, final OrmClassTool<BEAN> ormClassTool, final SaveOrUpdateType saveOrUpdateType) {
        if (cascade) {
            List<RelationInnerFK<BEAN, ?>> relations = ormClassTool.getClassMap().getInnerRelations();
            for (RelationInnerFK<BEAN, ?> relationInnerFk : relations) {
                RelationInnerFK<BEAN, RELATION> relationField = (RelationInnerFK<BEAN, RELATION>) relationInnerFk;
                RELATION innerBean = relationField.getGetManipulator().getValue(bean);
                RELATION saved = saveOrUpdateType.getStrategy().now(serviceCatalog.getSession(), innerBean, cascade, relationInnerFk.getCascadeInfo());
                relationField.getSetManipulator().setValue(bean, saved);
            }
        }
    }

    protected final <RELATION, BEAN> void saveOrUpdateOuterRelations(final boolean cascade, final BEAN bean, final OrmClassTool<BEAN> ormClassTool, final SaveOrUpdateType saveOrUpdateType) {
        if (cascade) {
            List<RelationOuterFK<BEAN, ?, ?>> relations = ormClassTool.getClassMap().getOuterRelations();
            if (!relations.isEmpty()) {
                String[] pks = ormClassTool.getClassMap().getPrimaryKeyColumnJavaNames();
                Object beanPrimaryKeyValue = ormClassTool.getOrmPersistor().getPropertyValues(pks, bean)[0];
                for (RelationOuterFK<BEAN,? ,?> relation : relations) {

                    ClassField<RELATION, Object> relationClassField = (ClassField<RELATION, Object>) relation.getRelationClassField();
                    Object saved = null;
                    if (relation.isOneToMany()) {
                        Collection<RELATION> relatedBeans = (Collection<RELATION>) relation.getGetManipulator().getValue(bean);
                        for (RELATION relatedBean : relatedBeans) {
                            relationClassField.getSetManipulator().setValue(relatedBean, beanPrimaryKeyValue);
                        }
                        saved = saveOrUpdateType.getStrategy().now(serviceCatalog.getSession(), relatedBeans, cascade, relation.getCascadeInfo());
                    } else {
                        RELATION relatedBean = (RELATION) relation.getGetManipulator().getValue(bean);
                        if (relatedBean!=null) {
                            relationClassField.getSetManipulator().setValue(relatedBean, beanPrimaryKeyValue);
                            saved = saveOrUpdateType.getStrategy().now(serviceCatalog.getSession(), relatedBean, cascade, relation.getCascadeInfo());
                        }
                    }
                    (( SetManipulator<BEAN, Object>) relation.getSetManipulator()).setValue(bean, saved);
                }
            }
        }
    }

    private <BEAN> String generateSaveQuery(final boolean useGenerator, final ClassMap<BEAN> classMap) {
        final StringBuilder builder = new StringBuilder("INSERT INTO "); //$NON-NLS-1$
        builder.append(classMap.getTableInfo().getTableNameWithSchema());
        builder.append(" ("); //$NON-NLS-1$
        builder.append( columnToCommaSepareted( "" , classMap.getAllColumnJavaNames(), !useGenerator, classMap )); //$NON-NLS-1$
        builder.append(") VALUES ("); //$NON-NLS-1$
        builder.append( questionCommaSepareted( classMap.getAllColumnJavaNames(), !useGenerator, classMap ));
        builder.append(") "); //$NON-NLS-1$
        return builder.toString();
    }

    private <BEAN> String questionCommaSepareted(final String[] fieldNames, final boolean ignoreGenerators, final ClassMap<BEAN> classMap) {
        final StringBuilder builder = new StringBuilder();
        final int length = fieldNames.length;
        boolean generatedKey = false;
        if (length > 0) {
            for (int i=0; i<(length-1) ; i++) {
                ClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[i]);
                final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, serviceCatalog.getDbProfile(), ignoreGenerators );
                generatedKey = generatedKey || columnValueGenerator.isAutoGenerated();
                final String queryParameter = columnValueGenerator.insertQueryParameter( "?"); //$NON-NLS-1$
                if (queryParameter.length()>0) {
                    builder.append( queryParameter );
                    builder.append(", "); //$NON-NLS-1$
                }
            }
            ClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[fieldNames.length-1]);
            final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, serviceCatalog.getDbProfile(), ignoreGenerators );
            generatedKey = generatedKey || columnValueGenerator.isAutoGenerated();
            builder.append( columnValueGenerator.insertQueryParameter( "?") ); //$NON-NLS-1$
        }
        return builder.toString();
    }

    private <BEAN> String columnToCommaSepareted(final String prefix, final String[] fieldNames, final boolean ignoreGenerators, final ClassMap<BEAN> classMap) {
        final StringBuilder builder = new StringBuilder();
        if (fieldNames.length > 0) {
            for (int i=0; i<(fieldNames.length-1) ; i++) {
                ClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[i]);
                final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, serviceCatalog.getDbProfile(), ignoreGenerators );
                final String queryParameter = columnValueGenerator.insertColumn(prefix + classField.getColumnInfo().getDBColumnName());
                if (queryParameter.length()>0) {
                    builder.append( queryParameter );
                    builder.append(", "); //$NON-NLS-1$
                }
            }
            ClassField<BEAN, ?> classField = classMap.getClassFieldByJavaName(fieldNames[fieldNames.length-1]);
            final AColumnValueGenerator columnValueGenerator = ColumnValueGeneratorFactory.getColumnValueGenerator( classField, serviceCatalog.getDbProfile(), ignoreGenerators );
            builder.append( columnValueGenerator.insertColumn( prefix + classField.getColumnInfo().getDBColumnName()) );
        }

        return builder.toString();
    }
}
